@@ -1,19 +1,24 @@
 from __future__ import annotations
 
-from .client import Client, create_config_file
+from . import errors, utils
+from .app import Application
+from .client import Client
 from .data import (
     AppData,
-    BackupData,
+    Backup,
+    BackupInfo,
+    DeployData,
+    DNSRecord,
+    DomainAnalytics,
     FileInfo,
     LogsData,
     PlanData,
-    StatisticsData,
+    ResumedStatus,
     StatusData,
     UploadData,
     UserData,
 )
 from .file import File
 from .http.endpoints import Endpoint
-from . import errors
 
-__version__ = '3.3.0'
+__version__ = '3.7.0'
