@@ -1,24 +1,27 @@
 from __future__ import annotations
 
-from abc import ABC
+from functools import wraps
 from io import BytesIO
-from typing import TYPE_CHECKING, Callable
+from typing import TYPE_CHECKING, Any, Callable
 
-from pydantic import PositiveInt
+from squarecloud import errors
 
+from ._internal.decorators import validate
 from .data import (
     AppData,
-    BackupData,
+    Backup,
+    BackupInfo,
     DeployData,
+    DNSRecord,
     DomainAnalytics,
     FileInfo,
     LogsData,
     StatusData,
 )
-from .errors import SquareException
 from .file import File
 from .http import Endpoint, HTTPClient, Response
-from .listeners import CaptureListenerManager
+from .listeners import Listener, ListenerConfig
+from .listeners.capture_listener import CaptureListenerManager
 
 # avoid circular imports
 if TYPE_CHECKING:
@@ -40,20 +43,20 @@ class AppCache:
         attributes.
 
 
-        :return: The instance of the class
-        :rtype: None
+        :return: None
         """
         self._status: StatusData | None = None
         self._logs: LogsData | None = None
-        self._backup: BackupData | None = None
+        self._backup: Backup | None = None
         self._app_data: AppData | None = None
 
     @property
     def status(self) -> StatusData:
         """
-        The status function returns the status of the application.
+        The status method is a property that returns the cached StatusData of
+        the application.
 
-        :return: The status of the application.
+        :return: A StatusData object.
         :rtype: StatusData
         """
         return self._status
@@ -61,7 +64,8 @@ class AppCache:
     @property
     def logs(self) -> LogsData:
         """
-        The logs function the logs of the application.
+        The logs method is a property that returns the cached LogsData of
+        the application.
 
         :return: The logs of your application
         :rtype: LogsData
@@ -69,19 +73,21 @@ class AppCache:
         return self._logs
 
     @property
-    def backup(self) -> BackupData:
+    def backup(self) -> Backup:
         """
-        The backup function is used to create a backup of the application.
+        The backup method is a property that returns the cached Backup of
+        the application.
 
         :return: The value of the _backup attribute
-        :rtype: BackupData
+        :rtype: Backup
         """
         return self._backup
 
     @property
     def app_data(self) -> AppData:
         """
-        The app_data function is a property that returns the AppData object.
+        The app_data method is a property that returns the cached AppData
+        object
 
         :return: The data from the app_data
         :rtype: AppData
@@ -90,12 +96,11 @@ class AppCache:
 
     def clear(self):
         """
-        The clear function is used to clear the status, logs, backup and data
+        The clear method is used to clear the status, logs, backup and data
         variables.
 
         :param self: Refer to the class instance
         :return: None
-        :rtype: None
         """
         self._status = None
         self._logs = None
@@ -104,24 +109,22 @@ class AppCache:
 
     def update(self, *args):
         """
-        The update function is used to update the data of a given instance.
+        The update method is used to update the data of a given instance.
         It takes in an arbitrary number of arguments, and updates the
         corresponding data if it is one of the following types:
-        StatusData, LogsData, BackupData or AppData.
+        StatusData, LogsData, Backup or AppData.
         If any other type is provided as an argument to this function,
         a SquareException will be raised.
 
-        :param self: Refer to the class instance
         :param args: Pass a variable number of arguments to a function
         :return: None
-        :rtype: None
         """
         for arg in args:
             if isinstance(arg, StatusData):
                 self._status = arg
             elif isinstance(arg, LogsData):
                 self._logs = arg
-            elif isinstance(arg, BackupData):
+            elif isinstance(arg, Backup):
                 self._backup = arg
             elif isinstance(arg, AppData):
                 self._app_data = arg
@@ -131,20 +134,16 @@ class AppCache:
                     for i in [
                         StatusData,
                         LogsData,
-                        BackupData,
+                        Backup,
                         AppData,
                     ]
                 ]
-                raise SquareException(
+                raise errors.SquareException(
                     f'you must provide stats of the following types:\n{types}'
                 )
 
 
-class AbstractApplication(ABC):
-    """Abstract application class"""
-
-
-class Application:
+class Application(CaptureListenerManager):
     """Represents an application"""
 
     __slots__ = [
@@ -154,12 +153,13 @@ class Application:
         '_data',
         'cache',
         '_id',
-        '_tag',
+        '_name',
         '_desc',
+        '_domain',
+        '_custom',
         '_ram',
         '_lang',
         '_cluster',
-        '_isWebsite',
         'always_avoid_listeners',
     ]
 
@@ -168,67 +168,68 @@ class Application:
         client: 'Client',
         http: HTTPClient,
         id: str,
-        tag: str,
-        ram: PositiveInt,
+        name: str,
+        ram: int,
         lang: str,
         cluster: str,
-        isWebsite: bool,
+        domain: str | None,
+        custom: str | None,
         desc: str | None = None,
     ) -> None:
         """
-        The `__init__` function is called when the class is instantiated.
+        The `__init__` method is called when the class is instantiated.
         It sets up all the attributes that are passed in as arguments,
         and does any other initialization your class needs before It's ready
         for use.
 
 
-        :param self: Refer to the class instance.
-        :param client: 'Client': Store a reference to the client that created
+        :param client: Store a reference to the client that created
         this app.
-        :param http: HTTPClient: Make http requests to the api.
-        :param id: str: The id of the app.
-        :param tag: str: The tag of the app.
-        :param ram: PositiveInt: The amount of ram that is allocated.
-        :param lang: str: The programming language of the app.
-        :param cluster: str: The cluster that the app is hosted on
-        :param isWebsite: bool: Whether if the app is a website
-        :param desc: str | None: Define the description of the app
+        :param http: Store a reference to the HTTPClient
+        :param id: The application id
+        :param name: The tag of the app
+        :param ram: The amount of ram that is allocated
+        :param lang: The programming language of the app
+        :param cluster: The cluster that the app is hosted on
+        :param desc: Define the description of the app
+        :param domain: Define the domain of the app
+        :param custom: Define the custom domain of the app
 
         :return: None
-        :rtype: None
         """
         self._id: str = id
-        self._tag: str = tag
+        self._name: str = name
+        self._domain: str | None = domain
+        self._custom: str | None = custom
         self._desc: str | None = desc
-        self._ram: PositiveInt = ram
+        self._ram: int = ram
         self._lang: str = lang
         self._cluster: str = cluster
-        self._isWebsite: bool = isWebsite
         self._client: 'Client' = client
         self._http: HTTPClient = http
         self._listener: CaptureListenerManager = CaptureListenerManager()
         self.cache: AppCache = AppCache()
         self.always_avoid_listeners: bool = False
+        super().__init__()
 
     def __repr__(self) -> str:
         """
-        The __repr__ function is used to create a string representation of an
+        The `__repr__` method is used to create a string representation of an
         object.
         This is useful for debugging, logging and other instances where you
         would want a string representation of the object.
         The __repr__ function should return a string that would make sense to
         someone looking at the results in the interactive interpreter.
 
-
-        :param self: Refer to the class instance
         :return: The class name, tag and id of the element
+        :rtype: str
         """
-        return f'<{self.__class__.__name__} tag={self.tag} id={self.id}>'
+        return f'<{self.__class__.__name__} tag={self.name} id={self.id}>'
 
     @property
     def client(self) -> 'Client':
         """
-        The client function is a property that returns the client object.
+        The client method is a property that returns the client object.
 
         :return: The client instance
         :rtype: Client
@@ -239,7 +240,7 @@ class Application:
     def id(self) -> str:
         """
         The id function is a property that returns
-         the id of the application.
+        the id of the application.
 
         :return: The id of the application
         :rtype: str
@@ -247,14 +248,14 @@ class Application:
         return self._id
 
     @property
-    def tag(self) -> str:
+    def name(self) -> str:
         """
         The tag function is a property that returns the application tag.
 
         :return: The tag of the application
         :rtype: str
         """
-        return self._tag
+        return self._name
 
     @property
     def desc(self) -> str | None:
@@ -279,9 +280,7 @@ class Application:
         return self._ram
 
     @property
-    def lang(
-        self,
-    ) -> str:
+    def lang(self) -> str:
         """
         The lang function is a property that returns the application's
         programing language.
@@ -292,9 +291,7 @@ class Application:
         return self._lang
 
     @property
-    def cluster(
-        self,
-    ) -> str:
+    def cluster(self) -> str:
         """
         The cluster function is a property that returns the
         cluster that the application is
@@ -307,19 +304,80 @@ class Application:
         return self._cluster
 
     @property
-    def is_website(self) -> bool:
+    def domain(self) -> str | None:
         """
-        The is_website function is a property that returns a boolean value
-        indicating whether th application is a website.
+        The domain function is a property that returns the
+        application's domain.
 
-        :return: A boolean value, true or false
-        :rtype: bool
-        """
-        return self._isWebsite
 
-    def capture(self, endpoint: Endpoint) -> Callable:
+        :return: The application's domain
+        :rtype: str | None
         """
-        The capture function is a decorator that can be used to add a function
+        return self._domain
+
+    @property
+    def custom(self) -> str | None:
+        """
+        The custom function is a property that returns the
+        application's custom domain.
+
+
+        :return: The application's domain
+        :rtype: str | None
+        """
+        return self._custom
+
+    @staticmethod
+    def _notify_listener(endpoint: Endpoint):
+        """
+        The _notify_listener function is a decorator that call a listener after
+        the decorated coroutine is called
+
+        :param endpoint: the endpoint for witch the listener will fetch
+        :return: a callable
+        """
+
+        def wrapper(func: Callable):
+            @wraps(func)
+            async def decorator(self: Application, *args, **kwargs) -> Any:
+                result = await func(self, *args, **kwargs)
+                avoid_listener = kwargs.pop('avoid_listener', False)
+                if not (avoid_listener or self.always_avoid_listeners):
+                    await self.notify(
+                        endpoint=endpoint,
+                        before=self.cache.app_data,
+                        after=result,
+                        extra=kwargs.get('extra'),
+                    )
+                return result
+
+            return decorator
+
+        return wrapper
+
+    @staticmethod
+    def _update_cache(func: Callable):
+        """
+        This is a decorator checks whether the kwargs `update_cache` in the
+        decorated coroutine is not False, and updates the application cache
+
+        :param func:
+        :return: a callable
+        """
+
+        @wraps(func)
+        async def wrapper(self, *args, **kwargs):
+            update_cache = kwargs.pop('update_cache', True)
+            result = await func(self, *args, **kwargs)
+            if update_cache:
+                self.cache.update(result)
+            return result
+
+        return wrapper
+
+    def capture(self, endpoint: Endpoint, **kwargs) -> Callable:
+        """
+        The capture function is a decorator that can be used to add a callable
         to be called when a request is made to the specified endpoint.
 
         :param self: Refer to the class instance.
@@ -328,14 +386,10 @@ class Application:
         :return: A decorator
         :rtype: Callable
         """
-        allowed_endpoints: tuple[Endpoint, Endpoint, Endpoint, Endpoint] = (
-            Endpoint.logs(),
-            Endpoint.app_status(),
-            Endpoint.backup(),
-            Endpoint.app_data(),
-        )
 
-        def wrapper(func) -> None:
+        def wrapper(
+            call: Callable[[Endpoint, Endpoint], Any]
+        ) -> Callable[[Endpoint, Endpoint], Any]:
             """
             The wrapper function is a decorator that takes in the endpoint as
             an argument.
@@ -345,116 +399,80 @@ class Application:
             with the function passed to wrapper().
             Otherwise, it raises another SquareException.
 
-            :param func: Pass the function to be wrapped
+            :param call: Pass the function to be wrapped
             :return: The wrapper function itself
             :rtype: None
+            :raises InvalidListener: Raised if the endpoint is already
+            registered
             """
-            if endpoint not in allowed_endpoints:
-                raise SquareException(
-                    f'the endpoint to capture must be {allowed_endpoints}'
-                )
-
-            if self._listener.get_capture_listener(endpoint) is None:
-                return self._listener.add_capture_listener(endpoint, func)
-            raise SquareException(
-                f'Already exists an capture_listener for {endpoint}'
-                f'{self._listener.get_capture_listener(endpoint)}'
+            for key, value in kwargs.items():
+                if key not in ListenerConfig.__annotations__:
+                    raise ValueError(
+                        f'Invalid listener configuration: "{key}={value}"'
+                    )
+            config = ListenerConfig(**kwargs)
+            listener = Listener(
+                app=self,
+                client=self.client,
+                endpoint=endpoint,
+                callback=call,
+                config=config,
             )
+            self.include_listener(listener)
+            return call
 
         return wrapper
 
-    async def data(
-        self, avoid_listener: bool = None, update_cache: bool = True
-    ) -> AppData:
+    @_update_cache
+    @_notify_listener(Endpoint.app_data())
+    async def data(self, *_args, **__kwargs) -> AppData:
         """
-        The data function is used to retrieve the data of an app.
+        The data method is used to retrieve the data of this app.
 
-        :param update_cache: if True, update the application cache
-        :param avoid_listener: whether the capture listener will be called
         :param self: Refer to the class instance
         :return: A AppData object
         :rtype: AppData
         """
         app_data: AppData = await self.client.app_data(self.id)
-        avoid_listener = avoid_listener or self.always_avoid_listeners
-        if not avoid_listener:
-            endpoint: Endpoint = Endpoint.app_data()
-            await self._listener.on_capture(
-                endpoint=endpoint,
-                before=self.cache.app_data,
-                after=app_data,
-            )
-        if update_cache:
-            self.cache.update(app_data)
         return app_data
 
-    async def logs(
-        self, avoid_listener: bool = None, update_cache: bool = True
-    ) -> LogsData:
+    @_update_cache
+    @_notify_listener(Endpoint.logs())
+    async def logs(self, *_args, **__kwargs) -> LogsData:
         """
-        The logs function is used to get the application's logs.
+        The logs method is used to get the application's logs.
 
         :param self: Refer to the class instance
-        :param update_cache: if True, update the application cache
-        :param avoid_listener: whether the capture listener will be called
         :return: A LogsData object
         :rtype: LogsData
         """
         logs: LogsData = await self.client.get_logs(self.id)
-        avoid_listener = avoid_listener or self.always_avoid_listeners
-        if not avoid_listener:
-            endpoint: Endpoint = Endpoint.logs()
-            await self._listener.on_capture(
-                endpoint=endpoint, before=self.cache.logs, after=logs
-            )
-        if update_cache:
-            self.cache.update(logs)
         return logs
 
-    async def status(
-        self, avoid_listener: bool = None, update_cache: bool = True
-    ) -> StatusData:
+    @_update_cache
+    @_notify_listener(Endpoint.app_status())
+    async def status(self, *_args, **__kwargs) -> StatusData:
         """
         The status function returns the status of an application.
 
         :param self: Refer to the class instance
-        :param update_cache: if True, update the application cache
-        :param avoid_listener: whether the capture listener will be called
         :return: A StatusData object
         :rtype: StatusData
         """
         status: StatusData = await self.client.app_status(self.id)
-        avoid_listener = avoid_listener or self.always_avoid_listeners
-        if not avoid_listener:
-            endpoint: Endpoint = Endpoint.app_status()
-            await self._listener.on_capture(
-                endpoint=endpoint, before=self.cache.status, after=status
-            )
-        if update_cache:
-            self.cache.update(status)
         return status
 
-    async def backup(
-        self, avoid_listener: bool = None, update_cache: bool = True
-    ) -> BackupData:
+    @_update_cache
+    @_notify_listener(Endpoint.backup())
+    async def backup(self, *_args, **__kwargs) -> Backup:
         """
         The backup function is used to create a backup of the application.
 
         :param self: Refer to the class instance
-        :param update_cache: if True, update the application cache
-        :param avoid_listener: whether the capture listener will be called
-        :return: A BackupData object
-        :rtype: BackupData
+        :return: A Backup object
+        :rtype: Backup
         """
-        backup: BackupData = await self.client.backup(self.id)
-        avoid_listener = avoid_listener or self.always_avoid_listeners
-        if not avoid_listener:
-            endpoint: Endpoint = Endpoint.backup()
-            await self._listener.on_capture(
-                endpoint=endpoint, before=self.cache.backup, after=backup
-            )
-        if update_cache:
-            self.cache.update(backup)
+        backup: Backup = await self.client.backup(self.id)
         return backup
 
     async def start(self) -> Response:
@@ -509,6 +527,7 @@ class Application:
         )
         return response
 
+    @validate
     async def commit(self, file: File) -> Response:
         """
         The commit function is used to commit the application.
@@ -524,6 +543,7 @@ class Application:
         )
         return response
 
+    @validate
     async def files_list(self, path: str) -> list[FileInfo]:
         """
         The files_list function returns a list of files and folders in the
@@ -541,6 +561,7 @@ class Application:
         )
         return response
 
+    @validate
     async def read_file(self, path: str) -> BytesIO:
         """
         The read_file function reads the contents of a file from an app.
@@ -555,6 +576,7 @@ class Application:
         )
         return response
 
+    @validate
     async def create_file(self, file: File, path: str) -> Response:
 
         """
@@ -574,6 +596,7 @@ class Application:
         )
         return response
 
+    @validate
     async def delete_file(self, path: str) -> Response:
         """
         The delete_file function deletes a file from the app.
@@ -603,6 +626,7 @@ class Application:
         )
         return response
 
+    @validate
     async def github_integration(self, access_token: str) -> str:
         """
         The create_github_integration function returns a webhook to integrate
@@ -625,8 +649,24 @@ class Application:
         )
         return analytics
 
+    @validate
     async def set_custom_domain(self, custom_domain: str):
         response: Response = await self.client.set_custom_domain(
             self.id, custom_domain, avoid_listener=True
         )
         return response
+
+    async def all_backups(self):
+        backups: list[BackupInfo] = await self.client.all_app_backups(self.id)
+        return backups
+
+    @validate
+    async def move_file(self, origin: str, dest: str):
+        return await self.client.move_app_file(self.id, origin, dest)
+
+    async def current_integration(self):
+        return await self.client.current_app_integration(self.id)
+
+    @_notify_listener(Endpoint.dns_records())
+    async def dns_records(self) -> list[DNSRecord]:
+        return await self.client.dns_records(self.id)
